"""Physical and internal objects.

This subpackage contains the definitions of many objects, with the purpose of
using them as type hints, but not only (at least for some of them).

Few related functions are also defined here.

"""
